
/*
 * PROGRAM STREAM DECAPSULATION
 *
 * Copyright (C) 1999  Thomas Mirlacher
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * The author may be reached as dent@cosy.sbg.ac.at, or
 * Thomas Mirlacher, Jakob-Haringerstr. 2, A-5020 Salzburg,
 * Austria
 *
 *------------------------------------------------------------
 *
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "dvb/ps.h"
#include "dvb/pes.h"
#include "plugin/decaps.h"

static int ps_decaps				(buf_t *buf);
static plugin_decaps_magic_t *ps_magic		(void);
static int ps_set_filter			(u_int filter_id, int val);

static struct ps_priv_struct {
	u_int filter[_BUF_MAX];
} priv;

static plugin_decaps_t decaps = {
	PLUGIN_ID_DECAPS,
	ps_decaps,	// decasulation
	ps_magic,	// autodetect decapsulation
	ps_set_filter,	// filtering
	NULL,		// read (inserted)
	NULL,		// write (inserted)
	2048,		// blocksize (inserted: this is a default)
	NULL		// private data
};

static buf_entry_t buf_entry; 

#define AUDIO_LPCM	0xA0
#define AUDIO_AC3	0x80
#define SPU_SUBTITLE	0x20

static char ps_file_magic[] = {
	0x00, 0x00, 0x01, 0xBA
};

static plugin_decaps_magic_t ps_magic_entry = {
	"ps;vob",
	4,
	ps_file_magic
};

/**
 *
 */

static inline char *_get_bytes_block (u_int len)
{
	if (!buf_entry.mem)
		buf_entry.mem = valloc (decaps.blocksize);// check return val
		//buf_entry.mem = malloc (decaps.blocksize);// check return val

	if (decaps.read (buf_entry.mem, decaps.blocksize) <= 0)
		return NULL;

	buf_entry.data = buf_entry.mem + len;

	return buf_entry.mem;
}

/**
 *
 */

static inline char *_get_bytes (u_int len)
{
	char *ptr = buf_entry.data;

	buf_entry.data += len;

	return ptr;
}


/**
 *
 */

static int ps_decaps (buf_t *buf)
{
	ps_hdr_t *ps;

//get new block
	if (!(ps = (ps_hdr_t *) _get_bytes_block (PS_HDR_LEN)))
		return -1;

	buf_entry.flags = BUF_FLAG_NONE;

// PS
	if (ps->stream_id == PS_STREAM_PACK_HDR) {
		ps_pack_t *ps_pack = (ps_pack_t *) _get_bytes (PS_PACK_LEN);

		buf_entry.clock = ps_pack->system_clock_reference_1;
		buf_entry.clock = buf_entry.clock<<2 | ps_pack->system_clock_reference_2;
		buf_entry.clock = buf_entry.clock<<8 | ps_pack->system_clock_reference_3;
		buf_entry.clock = buf_entry.clock<<5 | ps_pack->system_clock_reference_4;
		buf_entry.clock = buf_entry.clock<<2 | ps_pack->system_clock_reference_5;
		buf_entry.clock = buf_entry.clock<<8 | ps_pack->system_clock_reference_6;
		buf_entry.clock = buf_entry.clock<<5 | ps_pack->system_clock_reference_7;
	
		buf_entry.clock_ext = ps_pack->system_clock_reference_ext_1;
		buf_entry.clock_ext = buf_entry.clock_ext<<7 | ps_pack->system_clock_reference_ext_2;

		buf_entry.flags |= BUF_FLAG_CLOCK_VALID;

		_get_bytes (ps_pack->pack_stuffing_length);

		ps = (ps_hdr_t *) _get_bytes (PS_HDR_LEN);
	}

	if (ps->stream_id == PS_STREAM_SYSTEM_HDR) {
		ps_system_hdr_t *ps_system = (ps_system_hdr_t *) _get_bytes (2);
		u_int len;

		len = ps_system->header_length_hi<<8;
		len |= ps_system->header_length_lo;

		_get_bytes (len);

		ps = (ps_hdr_t *) _get_bytes (PS_HDR_LEN);
	}

// PES
	if ((ps->stream_id&STREAM_MASK_VIDEO) == STREAM_REST_VIDEO) {		// Video Stream;
		pes_hdr_t *hdr;
		pes_flag_t *flag;
		u_int pes_len;

		if (ps->stream_id != priv.filter[BUF_VIDEO])
			return 0;
	
		_get_bytes (-PS_HDR_LEN);
		hdr = (pes_hdr_t *) _get_bytes (PES_HDR_LEN);
		flag = (pes_flag_t *) _get_bytes (PES_FLAG_LEN);

		pes_len = hdr->pes_packet_length_hi<<8;
		pes_len |= hdr->pes_packet_length_lo;

		_get_bytes (flag->pes_header_data_length);

		buf_entry.data_len = pes_len - PES_FLAG_LEN - flag->pes_header_data_length;
		buf_entry.buf_id = BUF_VIDEO;
		decaps.write (buf, &buf_entry);
		buf_entry.mem = NULL;
	} else if (ps->stream_id == 0xBD) {	// Audio Stream (AC3)
		pes_hdr_t *hdr;
		pes_flag_t *flag;
		u_int pes_len;

		_get_bytes (-PS_HDR_LEN);
		hdr = (pes_hdr_t *) _get_bytes (PES_HDR_LEN);
		flag = (pes_flag_t *) _get_bytes (PES_FLAG_LEN);

		pes_len = hdr->pes_packet_length_hi<<8;
		pes_len |= hdr->pes_packet_length_lo;
	
		_get_bytes (flag->pes_header_data_length);

		if (*buf_entry.data == AUDIO_AC3 + priv.filter[BUF_AUDIO]) {
			if (priv.filter[BUF_AUDIO] < 0)
				return 0;
			_get_bytes (4);	

			buf_entry.data_len = pes_len - PES_FLAG_LEN - flag->pes_header_data_length - 4;
			buf_entry.buf_id = BUF_AUDIO;
			decaps.write (buf, &buf_entry);
			buf_entry.mem = NULL;
		} else if ((*buf_entry.data&0xf0) == SPU_SUBTITLE) { //0x20 + stream_id

// DENT: this has to be changed
			if (priv.filter[BUF_SUBPIC] == -1 ||
				(*buf_entry.data&0x0f) != priv.filter[BUF_SUBPIC])
				return 0;
	 
			_get_bytes (4);	

			buf_entry.data_len = pes_len - PES_FLAG_LEN - flag->pes_header_data_length - 4;
			buf_entry.buf_id = BUF_SUBPIC;
			decaps.write (buf, &buf_entry);
			buf_entry.mem = NULL;
		}
// else
//			fprintf(stderr,"unhandled audio stream: %x\n",*buf_entry.data);
	} else if (ps->stream_id == STREAM_PRIVATE_2) {	// SubPicture Stream
		pes_hdr_t *hdr;
		pes_flag_t *flag;
		u_int pes_len;

		_get_bytes (-PS_HDR_LEN);
		hdr = (pes_hdr_t *) _get_bytes (PES_HDR_LEN);
		flag = (pes_flag_t *) _get_bytes (PES_FLAG_LEN);

		pes_len = hdr->pes_packet_length_hi<<8;
		pes_len |= hdr->pes_packet_length_lo;
	
		_get_bytes (flag->pes_header_data_length);

		buf_entry.data_len = pes_len - PES_FLAG_LEN - flag->pes_header_data_length;
		buf_entry.buf_id = BUF_OTHER;
		decaps.write (buf, &buf_entry);
		buf_entry.mem = NULL;
	} else {				// Unhandled Stream
		fprintf (stderr, "%x ", ps->stream_id);
	}

	return 0;
}


/**
 * Autodetect decapsulation method
 */

static plugin_decaps_magic_t *ps_magic (void)
{
	return &ps_magic_entry;
}


/**
 * Filter streams
 */

static int ps_set_filter (u_int filter_id, int val)
{
	if (filter_id > _BUF_MAX)
		return -1;

	priv.filter[filter_id] = val;

	fprintf (stderr, "ps_set_filter called\n");
	return 0;
}


/**
 * Initial condition for this nice little thingy
 */

plugin_decaps_t *plugin_init (void *foo)
{
	buf_entry.mem_len = decaps.blocksize;

	fprintf (stderr, "decaps_init called\n");

//DENT just for debugging purposes
	priv.filter[BUF_VIDEO] = 0xE0;
	priv.filter[BUF_SUBPIC] = -1;	// subtitle substream id (0..31,-1=off)
	priv.filter[BUF_AUDIO] = 0x00;	// audio substream id (0...7, -1=off)

	return &decaps;
}


/**
 * Drop it
 */

void plugin_exit (void)
{
	fprintf (stderr, "decaps_close called\n");
}

