/*  Livid
 *
 * the playlist
 *
 * Copyright (C) 1999  Thomas Mirlacher, Guenter Bartsch
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#define __USE_LARGEFILE64

#include "playlist.h"

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "main.h"
#include "config_devices.h"
#include <nav/nav.h>
#include "panel_gui.h"
#include "messagebox.h"

struct navitem_info
{
  gchar *strInput;
  gchar *strTitle; // of DVD etc.
  gint   nTitle;
  gint   nChapter;
  gchar *strChapter;
};

void scroll_init (void);
extern struct player_global_struct player_global;
extern GtkWidget *widget_playlist;

/*
 * Global variables
 */

GtkWidget           *flatlist;
GtkTree             *tree;
GtkScrolledWindow   *flat_scroller;

gboolean             bFirstSet=FALSE; // TRUE if initial setup of 
                                      //player_global is done
struct navitem_info *pCurNavItem;
gboolean             bFlatlistMode=TRUE;
gint                 nSelectedRow=0; // in flatlist
gboolean             bSelectionInProgress = FALSE;


/*
 *
 */

void playlist_play_item (struct navitem_info *pItem)
{
  if (!pItem)
    {
      if (navGetStatus () != STATUS_STOP) 
	{
	  navSetStatus (STATUS_STOP);
	  panel_update_ui();
	}
      return;
    }


  fprintf (stderr, "\nNAVIGATE to title %d chapter %d on source %s\n",
	   pItem->nTitle, 
	   pItem->nChapter, 
	   pItem->strInput);

  /*
   * switch input source ? 
   */
 
  if ((player_global.strCurSource == NULL) 
      || (strcmp (player_global.strCurSource, pItem->strInput)!=0))
    {
      fprintf (stderr, "SWITCHING INPUT FROM %s to %s \n",player_global.strCurSource, pItem->strInput);

      // DENT: be sure to have player stopped before switching inputs   
      /*
       if (navGetStatus () != STATUS_STOP) 
         navSetStatus (STATUS_STOP);

	=> in nav itself now
      */

      navCloseInput (); // close previous input first

      if (navOpenInput (pItem->strInput) == DVD_PLAYER_NOT_FOUND) 
	{
	  char errstr[100];
	  sprintf (errstr, "Error switching input to %s", pItem->strInput);
	  show_message ("Error", errstr, MB_OK);
	  return;
	}
      if (player_global.strCurSource != NULL)
	free (player_global.strCurSource);

      player_global.strCurSource = g_strdup(pItem->strInput);

      if (player_global.strCurSourceName != NULL)
	free (player_global.strCurSourceName);

      player_global.strCurSourceName = g_strdup(pItem->strTitle);
    }

  
  PartPlay (pItem->nTitle, pItem->nChapter);

  pCurNavItem = pItem;

  player_global.nCurTitle   = pItem->nTitle;
  player_global.nCurChapter = pItem->nChapter;

  if (player_global.strCurChapter != NULL)
    free (player_global.strCurChapter);

  player_global.strCurChapter = g_strdup(pItem->strChapter);

  panel_update_ui();
}

/*
 * DnD stuff
 */


enum {
  TARGET_STRING,
  TARGET_ROOTWIN
};

static GtkTargetEntry target_table[] = {
  { "STRING",     0, TARGET_STRING },
  { "text/plain", 0, TARGET_STRING }
};

static guint n_targets = sizeof(target_table) / sizeof(target_table[0]);

void add_file (gchar *pathname, gint nRow)
{
  struct navitem_info *pNavItemInfo;
  gchar *texts[2] = { "F", pathname };

  if (nRow>0)
    nRow--;
      
  /*
   * generate navitem for file
   */
  
  pNavItemInfo = (struct navitem_info*) malloc (sizeof (struct navitem_info));
  pNavItemInfo->strInput  = g_strdup (pathname);
  pNavItemInfo->strTitle  = pNavItemInfo->strInput;
  pNavItemInfo->nTitle    = 0;
  pNavItemInfo->nChapter  = 0;
  pNavItemInfo->strChapter= "FILE-MODE";
  
  /*
   * add to flatlist
   */

  gtk_clist_insert( (GtkCList*) flatlist, nRow, texts);
  gtk_clist_set_row_data ( (GtkCList *) flatlist, nRow, pNavItemInfo);    
  texts[1] = g_strdup(pathname);
}

void flatlist_drag_data_received  (GtkWidget          *widget,
				   GdkDragContext     *context,
				   gint                x,
				   gint                y,
				   GtkSelectionData   *data,
				   guint               info,
				   guint               time)
{
  int nRow=0, nCol=0;
  gchar pathname[255];

  if ((data->length >= 0) && (data->format == 8))
    {
      
      if (sscanf (data->data, "file:%s", &pathname[0]) == 1)
	{
	  int yoffset ;
	  GtkAdjustment* pAdj;


	  pAdj = gtk_scrolled_window_get_vadjustment (flat_scroller);
	  yoffset = pAdj->value;
	  //   (GtkScrolledWindow *scrolled_window);

	  
	  gtk_clist_get_selection_info( (GtkCList *) widget,
					x, y+yoffset, &nRow, &nCol);

	  fprintf (stderr,"DROPPED: %s at %d \n",pathname, nRow);

	  add_file (pathname, nRow);
	  
	  gtk_drag_finish (context, TRUE, FALSE, time);
	}
    }
  
  gtk_drag_finish (context, FALSE, FALSE, time);
}


/**
 *
 */

void hide_playlist () 
{
  gtk_widget_hide (widget_playlist);
  
  player_global.bPLShowing = FALSE;
}

#define NAVITEM_CURRENT 0
#define NAVITEM_NEXT    1
#define NAVITEM_PREV    2

struct navitem_info *playlist_get_navitem (int nWich)
{
  if (!pCurNavItem)
    return NULL;

  if (bFlatlistMode)
    {
      gint nRow, nDesired=0;

      nRow = gtk_clist_find_row_from_data  ((GtkCList *) flatlist,
					    pCurNavItem);

      switch (nWich) 
	{
	case NAVITEM_CURRENT:
	  nDesired = nRow;
	  break;
	case NAVITEM_NEXT:
	  nDesired = nRow+1;
	  break;
	case NAVITEM_PREV:
	  nDesired = nRow-1;
	  break;
	}

      if (nDesired >= 0)
	{
	  struct navitem_info* pNavItemInfo;
	  
	  pNavItemInfo = (struct navitem_info *) gtk_clist_get_row_data ((GtkCList *) flatlist, nDesired);
      
	  if (pNavItemInfo)
	    {
	      bSelectionInProgress = TRUE;

	      gtk_clist_unselect_row ((GtkCList *) flatlist,
				    nRow, 0);

	      gtk_clist_select_row ((GtkCList *) flatlist,
				    nDesired, 0);

	      bSelectionInProgress = FALSE;
	    }  
	  return pNavItemInfo;

	}
      else
	return NULL;
    }
  else
    {
      GList *pChildren;
      /*
       * search device
       */

      pChildren =  tree->children;

      while (pChildren)
	{
	  gchar *str;

	  GtkTreeItem *pDeviceItem = (GtkTreeItem *) pChildren->data;
	  str = (char*) gtk_object_get_data(GTK_OBJECT(pDeviceItem), "device name");
	  
	  //	  fprintf (stderr, "DEVICE: %s =? %s\n",str,pCurNavItem->strInput);

	  if (strcmp (str, pCurNavItem->strInput) == 0)
	    {
	      /*
	       * get title
	       */

	      GtkTree *pTitleTree = (GtkTree *)pDeviceItem->subtree;
	      GtkTreeItem *pTitleItem = (GtkTreeItem *) g_list_nth(pTitleTree->children, pCurNavItem->nTitle)->data;

	      if (pTitleItem)
		{
		  GtkTree *pChapterTree = (GtkTree *)pTitleItem->subtree;
		  GtkTreeItem *pChapterItem;
		  int nDesired=0;

		  switch (nWich) 
		    {
		    case NAVITEM_CURRENT:
		      nDesired = pCurNavItem->nChapter;
		      break;
		    case NAVITEM_NEXT:
		      nDesired = pCurNavItem->nChapter+1;
		      break;
		    case NAVITEM_PREV:
		      nDesired = pCurNavItem->nChapter-1;
		      break;
		    }

		  //		  fprintf (stderr, "Desired : %d\n",nDesired);
		  
		  if (nDesired<0)
		    return NULL;

		  pChapterItem = (GtkTreeItem *) g_list_nth(pChapterTree->children, nDesired)->data;
		  if (pChapterItem)
		    {
		      //		      fprintf (stderr, "got Chapter item! :)))\n");
		      gtk_tree_select_child (pChapterTree, (GtkWidget *) pChapterItem);
		      return (struct navitem_info*) gtk_object_get_data(GTK_OBJECT(pChapterItem),"item_info");
		    }
		  
		}
	      else
		{
		  fprintf (stderr, "TREE: TitleItem not found!!\n");
		}
	    }
	  pChildren = g_list_next (pChildren);
	}

      //      fprintf (stderr, "Tree mode\n");
      return NULL;
    }

}

void playlist_play_next ()
{
  struct navitem_info *pNavItemInfo;

  pNavItemInfo = playlist_get_navitem (NAVITEM_NEXT);

  playlist_play_item (pNavItemInfo);

}

void playlist_play_prev ()
{
  struct navitem_info *pNavItemInfo;

  pNavItemInfo = playlist_get_navitem (NAVITEM_PREV);

  playlist_play_item (pNavItemInfo);
}

void playlist_play_current () 
{
  playlist_play_item (pCurNavItem);
}

void playlist_selection_made(GtkWidget *widget,
			     gint row,
			     gint column,
			     GdkEventButton *event,
			     gpointer data)
{
  struct navitem_info *pNavItemInfo;

  if (bSelectionInProgress)
    {
      //      fprintf (stderr,"COLLISION\n");
      return;
    }

  bSelectionInProgress = TRUE;

  nSelectedRow = row;
  bFlatlistMode = TRUE;

  if ((event->type==GDK_2BUTTON_PRESS)
      ||(event->type==GDK_3BUTTON_PRESS) ) {
    //    fprintf(stderr,"clicked on list more than one time\n");
 
    pNavItemInfo = (struct navitem_info *) gtk_clist_get_row_data ((GtkCList *) flatlist, row);


    playlist_play_item (pNavItemInfo);
  }

  bSelectionInProgress = FALSE;
}

// tree selection
void cb_itemsignal (GtkWidget *item, gchar *signame)
{
  struct navitem_info *pNavItemInfo;

  pNavItemInfo = (struct navitem_info*) gtk_object_get_data(GTK_OBJECT(item),
							    "item_info");

  bFlatlistMode = FALSE;
  playlist_play_item (pNavItemInfo);
}

void cb_chapter_change (void *data, int my_part)
{
  struct navitem_info *pNavItemInfo;

  int nTitle, nChapter;
  unsigned long int timeCode;
  int nlen;
  char str[256];

  fprintf (stderr,"\n\n CALLBACK CALLED \n\n");

  pNavItemInfo = playlist_get_navitem (NAVITEM_NEXT);

  if (pNavItemInfo)
    {

      GetRoot(&str[0], 256, &nlen);
      
      GetCurrentLocation(&nTitle, &nChapter, &timeCode); 

      nTitle = pNavItemInfo->nTitle;
      nChapter = my_part;
      
      fprintf (stderr,"\n\n CALLBACK: Current : %s T %d C %d\n", str, nTitle, nChapter);
      fprintf (stderr,"CALLBACK: Next    : %s T %d C %d\n", 
	       pNavItemInfo->strInput, 
	       pNavItemInfo->nTitle, 
	       pNavItemInfo->nChapter);
      
      if ((pNavItemInfo->nTitle != nTitle)
	  || (pNavItemInfo->nChapter != nChapter)
	  || (strcmp (pNavItemInfo->strInput, str) != 0))
	{

	  gdk_threads_enter();
	  playlist_play_item (pNavItemInfo);
	  gdk_threads_leave();
	  
	}
      else
	{
	  pCurNavItem = pNavItemInfo;

	  player_global.nCurTitle   = pNavItemInfo->nTitle;
	  player_global.nCurChapter = pNavItemInfo->nChapter;

	  if (player_global.strCurChapter != NULL)
	    free (player_global.strCurChapter);

	  player_global.strCurChapter = g_strdup(pNavItemInfo->strChapter);

	  gdk_threads_enter();
	  panel_update_ui();
	  gdk_threads_leave();

	}
    }
  else
    {
      if (navGetStatus () != STATUS_STOP) 
	{
	  navSetStatus (STATUS_STOP);
	  panel_update_ui();
	}
    }
}

void playlist_move_up (GtkButton *button,
		       gpointer user_data) 
{
  if (nSelectedRow>0)
    {
      gtk_clist_row_move ((GtkCList *) flatlist,
			  nSelectedRow, nSelectedRow-1);
      nSelectedRow--;
    }
}

void playlist_move_down (GtkButton *button,
			 gpointer user_data) 
{
  GtkCList *pList = (GtkCList *) flatlist;

  if (nSelectedRow<(pList->rows-1))
    {
      gtk_clist_row_move ((GtkCList *) flatlist,
			  nSelectedRow, nSelectedRow+1);
      nSelectedRow++;
    }
}

void playlist_remove (GtkButton *button,
		      gpointer user_data) 
{
  struct navitem_info* pNavItemInfo;
  GtkCList *pList = (GtkCList *) flatlist;

  if (nSelectedRow>=0)
    {
      pNavItemInfo = (struct navitem_info *) gtk_clist_get_row_data (pList, 
								     nSelectedRow);
      if (pNavItemInfo == pCurNavItem)
	  playlist_play_next();

      gtk_clist_remove (pList, nSelectedRow);

      // select same row (which is the next item now) if possible

      if (nSelectedRow<pList->rows)
	{
	  bSelectionInProgress = TRUE;

	  gtk_clist_select_row ((GtkCList *) flatlist,
				nSelectedRow, 0);

	  bSelectionInProgress = FALSE;
	}

    }
}

void playlist_add_file (GtkButton *button,
			gpointer user_data) 
{
}

void playlist_clear (GtkButton *button,
		     gpointer user_data) 
{
  if (bFlatlistMode)
    {
      pCurNavItem = NULL;
      player_global.strCurSourceName = "none";
      player_global.strCurChapter    = "";
      player_global.nCurChapter = 0;
      player_global.nCurTitle = 0;

      if (navGetStatus () != STATUS_STOP) 
	{
	  navSetStatus (STATUS_STOP);
	  panel_update_ui();
	}
    }

  gtk_clist_clear ((GtkCList*) flatlist);
}


GtkTree *scan_device (gchar *strDevName)
{
  GtkTree             *pSubTree, *pTitleTree=NULL;
  GtkWidget           *pItem;
  gint                 i, s, npos;
  nav_tree_t          *nav_tree;
  gchar               *strlist[3] = { NULL, NULL, NULL } ;
  gchar               *str;
  struct navitem_info *pNavItemInfo;
  GtkCList            *pList = (GtkCList *) flatlist;
  GList               *pChildren;

  gtk_widget_hide (GTK_WIDGET(tree));
  gtk_widget_hide (flatlist);


  /*
   * clear all entries of this device
   */

  // clear playlist
  fprintf (stderr, "Clear playlist \n");

  i=0;
  while (i<pList->rows)
    {
      pNavItemInfo = (struct navitem_info *) gtk_clist_get_row_data (pList,i); 

      if (strcmp(pNavItemInfo->strInput, strDevName)==0)
	gtk_clist_remove(pList,i);
      else
	i++;
    }

  // clear tree

  pChildren =  tree->children;

  while (pChildren)
    {
      gchar *str;

      GtkTreeItem *pDeviceItem = (GtkTreeItem *) pChildren->data;
      str = (char*) gtk_object_get_data(GTK_OBJECT(pDeviceItem), "device name");
	  
      if (strcmp (str, strDevName) == 0)
	{
	  // found it!

	  pTitleTree = GTK_TREE (pDeviceItem->subtree);

 	  if (pTitleTree) {
	    // replace

	    pTitleTree = (GtkTree *) gtk_tree_new();
	    
	    gtk_tree_item_set_subtree (GTK_TREE_ITEM(pDeviceItem), 
				       GTK_WIDGET(pTitleTree));

	  }
	}
      pChildren = g_list_next (pChildren);
    }


  /*
   * => sorry, but AFAIK this is still not working
   * DENT: it works for me (so i can just test DVD right now,
   *	but it seems to be ok)
   * DENT: but be aware: free the nav_tree !!!
   *
   */

  navCloseInput ();
  
  if (navOpenInput (strDevName) == DVD_PLAYER_NOT_FOUND)
    { 
      gchar errstr[100];
      sprintf (errstr, "Error opening device\n%s\nfor scanning", strDevName);
      gdk_threads_enter();
      show_message ("Error",errstr,MB_OK);
      gdk_threads_leave();

      
      pTitleTree = gtk_tree_new(); 
      pItem = gtk_tree_item_new_with_label ("error"); 
      gtk_tree_append (GTK_TREE(pTitleTree), pItem); 
      gtk_widget_show (pItem); 
      return pTitleTree; 
    }  


  nav_tree = navGetInfo ();
  
  if (!pTitleTree)
    pTitleTree = (GtkTree *) gtk_tree_new();

  for (i=0; i<nav_tree->num; i++) {
    GtkWidget *item;
    nav_subtree_t *nav_sub = nav_tree->title+i;
    
    pItem = gtk_tree_item_new_with_label (g_strdup(nav_sub->name));
    gtk_tree_append (GTK_TREE(pTitleTree), pItem);
    gtk_widget_show (pItem);

    pSubTree = (GtkTree*) gtk_tree_new();
    gtk_tree_item_set_subtree (GTK_TREE_ITEM(pItem), GTK_WIDGET(pSubTree));
    
    for (s=0; s<nav_sub->num; s++) {
      nav_tree_data_t *nav_data = nav_sub->sub+s;

      pNavItemInfo = (struct navitem_info*) malloc (sizeof (struct navitem_info));
      pNavItemInfo->strInput   = strDevName;
      pNavItemInfo->strTitle   = g_strdup (nav_tree->title->name);
      pNavItemInfo->nTitle     = i;
      pNavItemInfo->nChapter   = s;
      pNavItemInfo->strChapter = g_strdup(nav_data->name);

      if (!bFirstSet)
	{
	  player_global.nCurTitle     = i;
	  player_global.nCurChapter   = s;
	  player_global.strCurSourceName = g_strdup (nav_tree->title->name);
	  player_global.strCurSource  = g_strdup (strDevName);
	  player_global.strCurChapter = g_strdup(nav_data->name);
	  panel_update_ui();
	  pCurNavItem = pNavItemInfo;
	  bFirstSet = TRUE;
	}
      
      item = gtk_tree_item_new_with_label (g_strdup(nav_data->name));
      gtk_object_set_data(GTK_OBJECT(item), "item_info", pNavItemInfo); 


      gtk_tree_append (GTK_TREE(pSubTree), item);
      
      gtk_signal_connect (GTK_OBJECT(item), "select",
			  GTK_SIGNAL_FUNC(cb_itemsignal), "select");
      
      gtk_widget_show (item);

      /*
       * playlist entry
       */

      str = malloc (10);
      sprintf (str, "%d", i);

      strlist[0] = str;
      strlist[1] = g_strdup(nav_data->name);
      npos = gtk_clist_append( (GtkCList *) flatlist, (gchar **) &strlist);
      gtk_clist_set_row_data ( (GtkCList *) flatlist, npos, pNavItemInfo);

    }
  }

  gtk_widget_show (GTK_WIDGET(tree));
  gtk_widget_show (flatlist);

  return pTitleTree;
}

void playlist_scan_device (GtkButton *button,
			   gpointer user_data) 
{
  scan_device ((gchar *) user_data);
}

/*
 * go through all devices an collect playable items from them
 */

void scan_devices ()
{
  GtkWidget  *pItem;
  GList      *pDevList, *pDevCur;
  gchar      *strDevName;

  bFirstSet = FALSE;

  /*
   * TODO: clear Tree
   */

  pDevList = get_dev_list () ; // from config_devices

  /* 
   * now : scan devices
   */

  pDevCur = g_list_first (pDevList);

  while (pDevCur)
    {
      strDevName = (gchar *) pDevCur->data ;

      pItem = gtk_tree_item_new_with_label (strDevName);
      gtk_object_set_data(GTK_OBJECT(pItem), "device name", g_strdup(strDevName)); 
      gtk_tree_append (GTK_TREE(tree), pItem);
      gtk_widget_show (pItem);
    
      gtk_tree_item_set_subtree (GTK_TREE_ITEM(pItem), 
				 GTK_WIDGET(scan_device (strDevName)));

      pDevCur = g_list_next (pDevCur);
    }  

}

/**
 *
 */

GtkWidget* setup_playlist ()
{
  GtkWidget *win, *scrolled_win;
  GtkWidget *notebook, *label, *vbox, *hbox, *button;
  GList     *pDevList, *pDevCur;
  gchar     *strTitles[3] = { "Source", "Caption", NULL } ;
  
  //  gdk_threads_enter();

  win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_signal_connect (GTK_OBJECT(win), "delete_event",
		      GTK_SIGNAL_FUNC (hide_playlist), NULL);
  
  gtk_window_set_title(GTK_WINDOW(win), "DVD Navigator");
  

  /*
   * notebook for flatflist and tree 
   */

  notebook = gtk_notebook_new ();
  gtk_notebook_set_tab_pos (GTK_NOTEBOOK (notebook), GTK_POS_TOP);
  gtk_container_set_border_width (GTK_CONTAINER (notebook), 5);
  gtk_container_add (GTK_CONTAINER(win), notebook);
  gtk_widget_show(notebook);

  /*
   * the flatlist ("playlist view") 
   */

  hbox = gtk_hbox_new (FALSE, 2);
  gtk_widget_show (hbox);

  flat_scroller = GTK_SCROLLED_WINDOW (gtk_scrolled_window_new (NULL, NULL));
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (flat_scroller),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize (GTK_WIDGET(flat_scroller), 320, 200);
  gtk_widget_show (GTK_WIDGET(flat_scroller));
  
  flatlist = gtk_clist_new_with_titles (2,strTitles);
  gtk_widget_show (flatlist);

  
  gtk_signal_connect_object (GTK_OBJECT (flatlist), "select_row",
			     GTK_SIGNAL_FUNC(playlist_selection_made),
			     NULL);
  
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW(flat_scroller), flatlist);

  gtk_drag_dest_set (flatlist,
                     GTK_DEST_DEFAULT_ALL,
                     target_table, n_targets,
                     GDK_ACTION_COPY | GDK_ACTION_MOVE);
 
  gtk_signal_connect( GTK_OBJECT(flatlist), "drag_data_received",
                      GTK_SIGNAL_FUNC( flatlist_drag_data_received), NULL);

  gtk_box_pack_start (GTK_BOX (hbox), GTK_WIDGET(flat_scroller), 
		      TRUE, TRUE, 2);

  /*
   * flatlist-buttons
   */

  vbox = gtk_vbox_new (FALSE, 0);
  gtk_box_pack_start (GTK_BOX (hbox), vbox, FALSE, FALSE, 2);
  gtk_widget_show(vbox);

  button = gtk_button_new_with_label ("Up");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
 			     (GtkSignalFunc) playlist_move_up,
	 		     GTK_OBJECT (notebook));
  
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("Down");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
 			     (GtkSignalFunc) playlist_move_down,
	 		     GTK_OBJECT (notebook));

  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("Add File");
  
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) playlist_add_file,
			     GTK_OBJECT (notebook));
  
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("Remove");
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
			     (GtkSignalFunc) playlist_remove, 
			     GTK_OBJECT (notebook)); 
  
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);

  button = gtk_button_new_with_label ("Clear");
  
  gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
	 		     (GtkSignalFunc) playlist_clear, 
		 	     GTK_OBJECT (notebook)); 
  
  gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
  gtk_widget_show(button);

  /*
   * one scan button per device:
   */

  pDevList = get_dev_list () ; // from config_devices

  pDevCur = g_list_first (pDevList);

  while (pDevCur)
    {
      gchar  str[80];
      gchar *strDevName;
      strDevName = (gchar *) pDevCur->data ;

      fprintf (stderr, "Scan %s\n", strDevName);
      sprintf (str, "Scan %s", strDevName);

      button = gtk_button_new_with_label (str);
      
      gtk_signal_connect (GTK_OBJECT (button), "clicked",
			  (GtkSignalFunc) playlist_scan_device,
			  strDevName);
      

      gtk_box_pack_start (GTK_BOX (vbox), button, FALSE, FALSE, 2);
      gtk_widget_show(button);
      
      pDevCur = g_list_next (pDevCur);
    }

  // add flatlist-page

  label = gtk_label_new ("Playlist");
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), hbox, label);

  /*
   * the tree ("device view")
   */

  scrolled_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scrolled_win),
				  GTK_POLICY_AUTOMATIC,
				  GTK_POLICY_AUTOMATIC);
  gtk_widget_set_usize (scrolled_win, 200, 300);

  label = gtk_label_new ("Devices");
  gtk_notebook_append_page (GTK_NOTEBOOK (notebook), scrolled_win, label);

  gtk_widget_show (scrolled_win);
  
  tree = (GtkTree *) gtk_tree_new ();
  gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW(scrolled_win), 
					 (GtkWidget *) tree);
  gtk_tree_set_selection_mode (tree, GTK_SELECTION_SINGLE);
  gtk_tree_set_view_mode (tree, GTK_TREE_VIEW_LINE);
  
  gtk_widget_show ((GtkWidget *) tree);
  
  // get data to fill tree

  gdk_threads_enter();
  scan_devices ();
  gdk_threads_leave();

  navRegisterCb (CB_CHANGE, cb_chapter_change, NULL);

  return win;
}
