#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <string.h>

#include <plugin/input.h>
#include "input_vcd.h"

static int set_position(int min,int sec);
static int read_toc();
static int get_track_number(char m, char s, char f);
static int msf_to_lba (char m, char s, char f);
static void next_sector();

typedef struct {
	char sync	[12],
	header		[4],
	subheader	[8],
	data		[2324],
	spare		[4];
} cdsector;

static vcd_priv_t priv;

static int vcd_open (void);
static int vcd_close (void);
static int vcd_read (char *data, u_int len);
static int vcd_seek (u_long pos);


static plugin_input_t input = {
	PLUGIN_ID_INPUT,
	vcd_open,
	vcd_close,
	vcd_read,
	vcd_seek,
	CD_FRAMESIZE_RAW,	
	&priv
};


/**
 *
 */

static int vcd_open (void)
{
	return 0;
}


/**
 * Seek to a specific track
 */

static int vcd_seek (u_long pos)
{
	priv.m = priv.tocent[pos].cdte_addr.msf.minute;
	priv.s = priv.tocent[pos].cdte_addr.msf.second;
	priv.f = priv.tocent[pos].cdte_addr.msf.frame;
	priv.current_track = pos;
	priv.counter=0;

	return 0;
}


/**
 *
 */

static int vcd_read (char *buffer, u_int len) {
	struct cdrom_msf msf;
	int track;

	if (len != CD_FRAMESIZE_RAW)
		return -1;

	msf.cdmsf_min0 = priv.m;
	msf.cdmsf_sec0 = priv.s;
	msf.cdmsf_frame0 = priv.f;

	memcpy (buffer, &msf, sizeof (msf));

	if (ioctl(priv.fd, CDROMREADRAW, buffer)) {
		int t=get_track_number(priv.m,priv.s,priv.f);
		perror("cdromreadraw");
		if (t == priv.tochdr.cdth_trk1)
			return 0;

		vcd_seek (t);
		return vcd_read (buffer,len); // try again with next track
	}
	next_sector();

	if (priv.callb.chapter) {
		priv.counter++;
		if (priv.counter % 64 == 0) {
			track=get_track_number(priv.m,priv.s,priv.f)-1;
			if (track != priv.current_track) {
				priv.callb.chapter (priv.callb.chapter_data, track);
				priv.current_track=track;
			}
		}
	}

	return len;
}


/**
 *
 */

static int msf_to_lba (char m, char s, char f)
{
	return (((m * CD_SECS) + s) * CD_FRAMES + f) - CD_MSF_OFFSET;
}


/**
 *
 */

static int get_track_number (char m, char s, char f)
{
	int i;
	int lba2;
	int lba = msf_to_lba(m,s,f);

	for (i=priv.tochdr.cdth_trk0; i <= priv.tochdr.cdth_trk1; i++) {
		lba2 = msf_to_lba(priv.tocent[i-1].cdte_addr.msf.minute,priv.tocent[i-1].cdte_addr.msf.second,priv.tocent[i-1].cdte_addr.msf.frame);
		if (lba2>lba) return i-1;
	}

	return -1;
}


/**
 *
 */

static void next_sector()
{
	priv.f++;

	if (priv.f>=CD_FRAMES) {
		priv.f = 0;
		priv.s++;

		if (priv.s>=CD_SECS) {
			priv.s = 0;
			priv.m++;
		}
	}
}

  
/**
 *
 */

static int read_toc()
{
	int i;

  /* read toc header */
	if ( ioctl(priv.fd, CDROMREADTOCHDR, &priv.tochdr) == -1 ) {
		fprintf(stderr, "readvcd: ioctl cdromreadtochdr\n");
		return -1;
	}

  /* read individual tracks */
	for (i=priv.tochdr.cdth_trk0; i<=priv.tochdr.cdth_trk1; i++) {
		priv.tocent[i-1].cdte_track = i;
		priv.tocent[i-1].cdte_format = CDROM_MSF;
		if ( ioctl(priv.fd, CDROMREADTOCENTRY, &priv.tocent[i-1]) == -1 ) {
			fprintf(stderr, "readvcd: ioctl cdromreadtocentry\n");
			return -1;
		}
	}

  /* read the lead-out track */
	priv.tocent[priv.tochdr.cdth_trk1].cdte_track = CDROM_LEADOUT;
	priv.tocent[priv.tochdr.cdth_trk1].cdte_format = CDROM_MSF;

	if (ioctl(priv.fd, CDROMREADTOCENTRY, &priv.tocent[priv.tochdr.cdth_trk1]) == -1 ) {
		fprintf(stderr, "readvcd: ioctl cdromreadtocentry\n");
		return -1;
	}

	priv.total_tracks = priv.tochdr.cdth_trk1;

	return 0;
}


/**
 *
 */

static int set_position(int min,int sec) {
	if (min > priv.tocent[priv.tochdr.cdth_trk1].cdte_addr.msf.minute)
		return -1;

	if (min < 0 || sec < 0)
		return -1;

	priv.m=min;
	priv.s=sec;
	priv.f=0;

	return 0;
}


// MATS: please check this out, i've no idea if we can map this to
//	seek, or do we have to add an extra entry here?

#if 0
int VCDSetTime(int time) {
	return set_position(0,time/60,time%60);
}
  
/* TODO: fix track_end */
int VCDSetTrack(int track, int track_end) {
  return set_position(track+1,-1,-1);
}


    
int VCDGetTime() {
  return m*60+s;
      }
      
int VCDGetTotalTime() {
  return  priv.tocent[priv.tochdr.cdth_trk1].cdte_addr.msf.minute*60+priv.tocent[priv.tochdr.cdth_trk1].cdte_addr.msf.second;
      }
      
int VCDGetTrack() {
  return get_track_number(m,s,f)-1;
}
      
int VCDGetTotalTracks() {
  return priv.tochdr.cdth_trk1-1;
      }
      
      

#endif


/**
 *
 */

plugin_input_t *plugin_init (char *device)
{
	if ((priv.fd=open (device, O_RDONLY)) == -1) {
		perror (device);
		return NULL;
	}

	if (read_toc ())
		return NULL;

	if (vcd_seek (0))
		return NULL;

	return &input;
}


/**
 *
 */

void plugin_exit (void) {
	close (priv.fd);
}
