#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <dlfcn.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <netinet/in.h>

#include "udf/dvd_udf.h"
#include "ifo/libifo/ifo.h"
#include "plugin/plugin.h"
#include "plugin/nav.h"
#include "nav.h"

#define DVDDB

static int dvdOpen (plugin_t *plugin, char *path);
static int dvdPartPlayAutoStop (int title, int chapter, int parts2play);
static int dvdRegisterCb (u_int id, void (*cb) (void *, int ), void *data);
static int dvdUnregisterCb (u_int id);
static nav_tree_t *dvdGetInfo ();
static int _ifo_init ();
static int _get_cellpiece_index (void);
static int _set_first_block ();
static int get_next_block (char *data, u_int len);

#ifdef DVDDB
#include <dvddb/dvddb.h>
#endif

#define FILE_IFO "/VIDEO_TS/VIDEO_TS.IFO"

static struct {
	int audio;
	int subpic;

	int title;
	int chapter;

	long block;
	u_int title_start;

	u_int cell_index;
	u_int cell_vob_id;
	u_int cell_cell_id;

	u_int cellpiece_index;

	ifo_cell_addr_t *ptr_cell_addr;
	int fd_ifo;

	ifo_t *ifo;
	char *root;

	plugin_t *plugin;

	struct {
		void (*chapter) (void *data, int chapter);
		void *chapter_data;
	} callb;        // Callbacks
} dvd_priv;


static plugin_nav_t dvd_nav = {
	dvdOpen,
	NULL,
	dvdPartPlayAutoStop,
	dvdGetInfo,
	dvdRegisterCb,		// register callback
	dvdUnregisterCb,	// unregister callback
	&dvd_priv
};


/**
 * Initialize Navigation plugin.
 */

plugin_nav_t *plugin_init (void *foo)
{
fprintf (stderr, "dvdInit called\n");
	return &dvd_nav;
}


/**
 *
 */

void plugin_exit (void)
{
}


/**
 *
 */

static int dvdOpen (plugin_t *plugin, char *path)
{
	dvd_priv.root = path;
	dvd_priv.plugin = plugin;

fprintf (stderr, "dvdOpen called\n");

	if (pluginOpen (plugin, PLUGIN_ID_INPUT, DL_INPUT_UDF, path))
		return -1;

	if (pluginOpen (plugin, PLUGIN_ID_DECAPS, DL_DECAPS_PS, NULL))
		return -1;

	PLUGIN_OPS_DECAPS(plugin)->blocksize = 2048;
	PLUGIN_OPS_INPUT(plugin)->blocksize  = 2048; 
	PLUGIN_OPS_DECAPS(plugin)->read = get_next_block;

	if (!plugin->ops[PLUGIN_ID_CODEC]) {
		fprintf (stderr, "DEBUG: no codec registered!");
	} else 
		PLUGIN_OPS_DECAPS(plugin)->write = PLUGIN_OPS_CODEC(plugin)->read;

	if ((dvd_priv.fd_ifo = UDFOpenDisc (path)) < 0) {
		fprintf (stderr, "unable to open IFO %s\n", path);
		return -1;
	}

	if (dvd_priv.ifo) {
		ifoClose (dvd_priv.ifo);
		close (dvd_priv.fd_ifo);
	}

	_ifo_init ();

	PLUGIN_OPS_INPUT(plugin)->seek (dvd_priv.title_start);
	PLUGIN_OPS_INPUT(plugin)->open ();
	
	return DVD_PLAYER_OK;
}


/**
 *
 */

static int dvdPartPlayAutoStop (int title, int chapter, int parts2play)
{
	int num;
	char *ptr;
	char *pgci;
	plugin_t *plugin = dvd_priv.plugin;

	dvd_priv.title = title;
	dvd_priv.chapter = chapter;
//	dvd_priv.parts2play = parts2play;

	ifoGetPGCI ((ifo_hdr_t *) dvd_priv.ifo->data[ID_TITLE_PGCI], dvd_priv.title, &pgci),
	num = ifoGetProgramMap (pgci, &ptr);

	ptr += dvd_priv.chapter;
	dvd_priv.cell_index = *ptr - 1;

	if (((ifo_video_info_t *) dvd_priv.ifo->data[ID_MAT] + IFO_OFFSET_VIDEO)->system) {
		// VOB,50Hz,PAL
		plugin_codec_attr_video_t attr = {0, 0, -1, -1, 2};
		PLUGIN_OPS_CODEC(plugin)->set_attributes (BUF_VIDEO, &attr);
	} else {
		// VOB,60Hz,PAL60
		plugin_codec_attr_video_t attr = {0, 1, -1, -1, 7};
		PLUGIN_OPS_CODEC(plugin)->set_attributes (BUF_VIDEO, &attr);
	}

	{
		//AC3,16bit,48kHz
		plugin_codec_attr_audio_t attr = {0, 0, 1, 0,-1};
		PLUGIN_OPS_CODEC(plugin)->set_attributes (BUF_AUDIO, &attr);
	}

	return _set_first_block (plugin);
}


/**
 *
 */

static int dvdRegisterCb (u_int id, void (*cb) (void *, int ), void *data)
{
	fprintf (stderr, "UI registered callback\n");

	switch (id) {
	case CB_CHANGE:
		dvd_priv.callb.chapter = cb;
		dvd_priv.callb.chapter_data = data;
		break;
	default:
		return -1;
	}

	return 0;	
}


/**
 *
 */

static int dvdUnregisterCb (u_int id)
{
	fprintf (stderr, "UI unregistered callback %d\n", id);

	return 0;
}


/**
 *
 */

// "Title " + ?? + '\0'
#define TITLE_NAME_LEN (6 + 2 + 1)
// "Chapter " + ??? + '\0'
#define PART_NAME_LEN (7 + 3 + 1)

static nav_tree_t *dvdGetInfo ()
{
	ifo_ptt_t *ptt;
	nav_tree_t *tree;
        int i, s;
#ifdef DVDDB
	FILE *dvddb_file;
	char dvd_id[9];
#endif

	if (!(ptt = ifo_get_ptt (dvd_priv.ifo)))
		return NULL;

	tree = malloc (sizeof (nav_tree_t));
	tree->num = ptt->num;
	tree->title = malloc (sizeof (nav_subtree_t) * tree->num);

#ifdef DVDDB
	dvddbGetID (dvd_priv.root, dvd_id);

	if ((dvddb_file = dvddbOpen (dvd_id)) == NULL) {
		int answer;

		fprintf (stderr, "not found in local cache.\n");
		fprintf (stderr, "should i try to fetch it from the internet? (Y/n)");

		answer = getchar ();

		if (answer != 'n' && answer != 'N') {
			int sock;

			if ((sock = dvddbNetConnect ()) >= 0) {
				dvddbNetGet (sock, dvd_id);
				dvddbNetDisconnect (sock);

				dvddb_file = dvddbOpen (dvd_id);
			}
		}
	}
#endif

fprintf (stderr, "GetInfo: ptt_num: %d\n", ptt->num);

        for (i=0; i<ptt->num; i++) {
                ifo_ptt_sub_t *title = ptt->title+i;
		nav_subtree_t *sub_tree = tree->title+i;
		char *name;
#ifdef DVDDB
		char *name_ptr;
#endif

		sub_tree->sub = (nav_tree_data_t *) malloc (sizeof (nav_tree_data_t) * title->num);
		sub_tree->num = title->num;

#ifdef DVDDB
		if ((name_ptr = dvddbRead (dvddb_file))) {
			name = (char *) malloc (strlen (name_ptr) + 1);
			strcpy (name, name_ptr);
		} else
#endif
		{
			name = (char *) malloc (TITLE_NAME_LEN);
			snprintf (name, TITLE_NAME_LEN, "Title %d", i);
		}
		sub_tree->name = name;

#ifdef DVDDB
		dvddbRead (dvddb_file);		// read blank line after title
#endif
                for (s=0; s<title->num; s++) {
			nav_tree_data_t *data = sub_tree->sub+s;
			char *name;

#ifdef DVDDB
			if ((name = dvddbRead (dvddb_file))) {
				name = (char *) malloc (strlen (name_ptr) + 1);
				strcpy (name, name_ptr);
			} else
#endif
			{
				name  = (char *) malloc (PART_NAME_LEN);
				snprintf (name, PART_NAME_LEN, "Chapter %d", s);
			}
			data->name = name;
fprintf (stderr, "GetInfo: %d\t%s\n", s, name);
                }
#ifdef DVDDB
//		dvddbClose (dvddb_file);
#endif
        }

	return tree;
}

 
/**
 * Initialize Information files
 */

static int _ifo_init ()
{
	u_char *ptr;
	u_int offset;

	if (!(dvd_priv.title_start = UDFFindFile (FILE_IFO))) {
		fprintf (stderr, "unable to UDFFindFile %s\n", FILE_IFO);
		return -1;
	}

// opening VIDEO_TS.IFO
	if (!(dvd_priv.ifo = ifoOpen (dvd_priv.fd_ifo, dvd_priv.title_start * DVD_VIDEO_LB_LEN))) 
		return (-1);

// getting offset of first important ifo file
	ifoGetVMGPTT ((ifo_hdr_t *) dvd_priv.ifo->data[ID_PTT], (char **) &ptr);
	ptr += 8;

// DENT: check endian
        offset  = *(ptr+0) << 24;
        offset |= *(ptr+1) << 16;
        offset |= *(ptr+2) << 8;
        offset |= *(ptr+3);

	dvd_priv.title_start += offset;

// closing VIDEO_TS.IFO
	ifoClose (dvd_priv.ifo);

// opening VTS_??_0.IFO
	if (!(dvd_priv.ifo = ifoOpen (dvd_priv.fd_ifo, dvd_priv.title_start * DVD_VIDEO_LB_LEN))) {
		fprintf (stderr, "error initializing ifo.\n");
		exit (-1);
	}

	dvd_priv.title_start += ifoGetVOBStart (dvd_priv.ifo);

	ifoGetPGCI ((ifo_hdr_t *) dvd_priv.ifo->data[ID_TITLE_PGCI], dvd_priv.title, (char **) &ptr);

	ifoGetCellAddr ((char *) dvd_priv.ifo->data[ID_TITLE_CELL_ADDR], (char **) &dvd_priv.ptr_cell_addr);

	return 0;
}


/**
 *
 */

static int _get_cellpiece_index (void)
{
	ifo_cell_addr_t *cell_addr;
	ifo_pgc_cell_pos_t *cell_pos;
	char *pgci;
	char *clut;
	int index;
	plugin_codec_attr_spu_t attr;

	ifoGetPGCI ((ifo_hdr_t *) dvd_priv.ifo->data[ID_TITLE_PGCI], dvd_priv.title, &pgci),

// DENT a new CLUT starts in a new cell ... so move this out here
	ifoGetCLUT (pgci, (char **) &clut);
	attr.clut = clut;
	attr.num  = -1;
	PLUGIN_OPS_CODEC(dvd_priv.plugin)->set_attributes (BUF_SUBPIC, &attr);
// DENT end

	ifoGetCellPos (pgci, (char **) &cell_pos);
	
	cell_pos += dvd_priv.cell_index;
	dvd_priv.cell_vob_id = ntohs (cell_pos->vob_id);
	dvd_priv.cell_cell_id = ntohs (cell_pos->cell_id);

	cell_addr = dvd_priv.ptr_cell_addr;

// find cell piece
// DENT: we would be faster if we keep the last index, but we would need
//	some mechanism to reset the index

	for (index=0;;index++) {
		if ((ntohs (cell_addr->vob_id) == dvd_priv.cell_vob_id) &&
		    (cell_addr->cell_id == dvd_priv.cell_cell_id)) {
			break;
		}

		cell_addr ++;	
	}

	return index;
}


/**
 *
 */

static int _set_first_block ()
{
	ifo_cell_addr_t *cell_addr;
	plugin_t *plugin = dvd_priv.plugin;

	dvd_priv.cellpiece_index = _get_cellpiece_index ();

	cell_addr = dvd_priv.ptr_cell_addr;
	cell_addr += dvd_priv.cellpiece_index;

	dvd_priv.block = ntohl (cell_addr->start);

	PLUGIN_OPS_INPUT(plugin)->seek (dvd_priv.block + dvd_priv.title_start);

	return DVD_PLAYER_OK;
}


/**
 *
 */

static int get_next_block (char *data, u_int len)
{
	ifo_cell_addr_t *cell_addr;
	plugin_t *plugin = dvd_priv.plugin;

	cell_addr = dvd_priv.ptr_cell_addr;
	cell_addr += dvd_priv.cellpiece_index;

	if (dvd_priv.block > (ntohl (cell_addr->end))) { // next cell_piece
		cell_addr++;	// try next cellpiece

		if ((ntohs (cell_addr->vob_id) == dvd_priv.cell_vob_id) &&
		    (cell_addr->cell_id == dvd_priv.cell_cell_id)) { // ok

			dvd_priv.cellpiece_index++;
			dvd_priv.block = ntohl (cell_addr->start);
		} else {	// start next cell
			char *ptr;
			char *pgci;

			dvd_priv.cell_index++;

			ifoGetPGCI ((ifo_hdr_t *) dvd_priv.ifo->data[ID_TITLE_PGCI], dvd_priv.title, &pgci);

			if (dvd_priv.chapter < ifoGetProgramMap (pgci, &ptr))
				ptr += dvd_priv.chapter + 1;

			if (dvd_priv.cell_index == (*ptr - 1)) {
				dvd_priv.chapter ++;
				if (dvd_priv.callb.chapter)
					dvd_priv.callb.chapter (dvd_priv.callb.chapter_data, dvd_priv.chapter);

			}

			dvd_priv.cellpiece_index = _get_cellpiece_index ();

			cell_addr = dvd_priv.ptr_cell_addr;
			cell_addr += dvd_priv.cellpiece_index;

			if (cell_addr->start == (cell_addr-1)->start) {
				fprintf (stderr, "new MATRIX HACK!!!\n");
				dvd_priv.cellpiece_index++;
				cell_addr++;
			}

			dvd_priv.block = ntohl (cell_addr->start);
		}
	} else {
		dvd_priv.block++;
		return PLUGIN_OPS_INPUT(plugin)->read (data, len);
	}

	PLUGIN_OPS_INPUT(plugin)->seek ((dvd_priv.block++) + dvd_priv.title_start);
	return PLUGIN_OPS_INPUT(plugin)->read (data, len);
}
